from argparse import ArgumentParser
from dodo_commands.framework import Dodo, CommandError
from dodo_commands.framework.config import CommandPath
import os


def _args():
    parser = ArgumentParser(description="Creates a new Dodo command.")
    parser.add_argument('name')
    parser.add_argument(
        '--next-to',
        required=True,
        help='Create the new command at the location of this command')
    args = Dodo.parse_args(parser)
    return args


script = """
from argparse import ArgumentParser
from dodo_commands.framework import Dodo


def _args():
    parser = ArgumentParser()
    args = Dodo.parse_args(parser)
    return args


if Dodo.is_main(__name__, safe=True):
    args = _args()
    Dodo.runcmd([], cwd='.')
"""

if Dodo.is_main(__name__, safe=False):
    args = _args()
    dest_path = None
    command_path = CommandPath(Dodo.config)
    for item in command_path.items:
        script_path = os.path.join(item, args.next_to + ".py")
        if os.path.exists(script_path):
            dest_path = os.path.join(item, args.name + ".py")

    if not dest_path:
        raise CommandError("Script not found: %s" % args.next_to)

    if os.path.exists(dest_path):
        raise CommandError("Destination already exists: %s" % dest_path)

    with open(dest_path, "w") as f:
        f.write(script)
    print(dest_path)



from argparse import ArgumentParser
from dodo_commands.framework import Dodo


def _args():
    parser = ArgumentParser()
    parser.add_argument('what')
    parser.add_argument('where')
    args = Dodo.parse_args(parser)
    return args


if Dodo.is_main(__name__):
    args = _args()
    Dodo.runcmd(['grep', '-rnw', args.where, '-e', "'{}'".format(args.what)], )



# -*- coding: utf-8 -*-
"""Utilities."""
from six.moves import input as raw_input
import os
import sys
import re


def query_yes_no(question, default="yes"):
    """Ask a yes/no question via raw_input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = raw_input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")


class classproperty(object):  # noqa
    def __init__(self, f):
        self.f = f

    def __get__(self, obj, owner):
        """Return read-only property value."""
        return self.f(owner)


def remove_trailing_dashes(args):
    """Removes first -- item from args."""
    return args[1:] if args[:1] == ['--'] else args


def bordered(text):
    lines = text.splitlines()
    width = max(len(s) for s in lines)
    res = ['┌' + '─' * width + '┐']
    for s in lines:
        res.append('│' + (s + ' ' * width)[:width] + '│')
    res.append('└' + '─' * width + '┘')
    return '\n'.join(res)


def is_using_system_dodo():
    import dodo_commands
    dodo_commands_path = dodo_commands.__path__[0]
    return os.path.realpath(dodo_commands_path) == dodo_commands_path


try:
    import textwrap
    textwrap.indent
except AttributeError:  # undefined function (wasn't added until Python 3.3)

    def indent(text, amount, ch=' '):
        padding = amount * ch
        return ''.join(padding + line for line in text.splitlines(True))
else:

    def indent(text, amount, ch=' '):
        return textwrap.indent(text, amount * ch)


def filter_choices(all_choices, raw_choice):
    regexp = r"(\d)+(\-(\d)+)?,?"
    result = []
    span = [None, None]
    for choice in [x for x in re.finditer(regexp, raw_choice)]:
        if span[0] is None:
            span[0] = choice.span()[0]
        if span[1] is None or span[1] == choice.span()[0]:
            span[1] = choice.span()[1]
        from_index = int(choice.group(1)) - 1
        if from_index < 0:
            raise Exception("Invalid index")
        to_index = int(choice.group(3)) - 1 if choice.group(3) else from_index
        if to_index >= len(all_choices):
            raise Exception("Invalid index")
        for idx in range(from_index, to_index + 1):
            result.append(all_choices[idx])
    return result, span


def is_windows():
    return os.name == 'nt'


def symlink(source, link_name, target_is_directory=False):
    import os
    os_symlink = getattr(os, "symlink", None)
    if callable(os_symlink):
        os_symlink(source, link_name, target_is_directory)
    else:
        import ctypes
        csl = ctypes.windll.kernel32.CreateSymbolicLinkW
        csl.argtypes = (ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_uint32)
        csl.restype = ctypes.c_ubyte
        flags = 1 if os.path.isdir(source) else 0
        if csl(link_name, source, flags) == 0:
            raise ctypes.WinError()



from argparse import ArgumentParser
from dodo_commands.framework import Dodo, CommandError
import os
from six.moves import input as raw_input
from dodo_commands.framework.util import filter_choices
from plumbum.cmd import tmux

session_id = os.path.expandvars('$USER')


def _args():
    parser = ArgumentParser()
    parser.add_argument('--kill-session', action='store_true')
    args = Dodo.parse_args(parser)
    args.commands = Dodo.get_config('/TMUX/commands', [])
    return args


def _create_tmux_window():
    # Create tmux session
    Dodo.runcmd(['tmux', '-2', 'new-session', '-d', '-s', session_id], )

    # Create a tmux window
    Dodo.runcmd(
        ['tmux', 'new-window', '-t',
         '%s:1' % session_id, '-n', 'Logs'], )


if Dodo.is_main(__name__):
    args = _args()
    check_exists = Dodo.get_config('/TMUX/check_exists', '/')
    if not os.path.exists(check_exists):
        raise CommandError("Path %s does not exist" % check_exists)

    has_session = False
    try:
        sessions = tmux('ls')
        for session in sessions.split('\n'):
            has_session = has_session or session.startswith('%s:' % session_id)
    except:
        pass

    if has_session and args.kill_session:
        Dodo.runcmd(['tmux', 'kill-session', '-t', session_id], )

    if not has_session:
        _create_tmux_window()
        Dodo.runcmd(['tmux', 'send-keys', 'dodo tmux', 'C-m'], )
        # Attach to tmux session
        Dodo.runcmd(['tmux', '-2', 'attach-session', '-t', session_id], )
    else:
        while True:
            commands = []
            label_size = 0
            for label in args.commands:
                label_size = max(label_size, len(label))
            label_prefix = "%0" + str(label_size) + "s"

            print()
            for label in args.commands:
                for command in args.commands[label]:
                    commands.append(command)
                    format_string = "%02s [" + label_prefix + "] - %s"
                    print(format_string % (str(len(commands)), label, command))

            raw_choice = raw_input(
                '\nSelect one or more commands (e.g. 1,3-4), or type a command: '
            )
            selected_commands, span = filter_choices(commands, raw_choice)
            if span != [0, len(raw_choice)]:
                selected_commands = [raw_choice]

            for idx, command in enumerate(selected_commands):
                Dodo.runcmd(['tmux', 'split-window', '-v'], )
                Dodo.runcmd(['tmux', 'send-keys', command, 'C-m'], )

            # Set default window
            tmux('select-pane', '-t', '0')
            tmux('select-layout', 'tile')



"""Pauses the execution."""
import time
from dodo_commands.framework import Dodo


class Decorator:  # noqa
    def add_arguments(self, parser):  # noqa
        parser.add_argument(
            '--pause-ms',
            type=int,
            help="Pause in milliseconds before continuing")

    def modify_args(self, root_node, cwd):  # noqa
        if Dodo.args.pause_ms:
            time.sleep(Dodo.args.pause_ms / 1000)
        return root_node, cwd



import os
from plumbum import local


def _is_windows():
    return os.name == 'nt'


def _ext():
    return '.exe' if _is_windows() else ''


class Paths:
    def home_dir(self):
        return os.path.expanduser('~')

    def global_config_dir(self):
        return os.path.join(self.home_dir(), '.dodo_commands')

    def global_config_filename(self):
        return os.path.join(self.global_config_dir(), 'config')

    def default_commands_dir(self):
        return os.path.join(self.global_config_dir(), 'default_commands')

    def virtual_env_dir(self):
        return os.path.join(self.project_dir(), 'dodo_commands', 'env')

    def virtual_env_bin_dir(self):
        return os.path.join(self.virtual_env_dir(), 'Scripts'
                            if _is_windows() else 'bin')

    def pip(self):
        return os.path.join(self.virtual_env_bin_dir(), 'pip' + _ext())

    def site_packages_dir(self):
        python = local[os.path.join(self.virtual_env_bin_dir(),
                                    "python" + _ext())]

        result = python(
            "-c", "from distutils.sysconfig import get_python_lib; " +
            "print(get_python_lib())")
        while result[-1] in ['\n', '\r']:
            result = result[:-1]
        return result

    def project_dir(self):
        """Return the root dir of the current project."""
        return os.environ['DODO_COMMANDS_PROJECT_DIR']

    def res_dir(self):
        return os.path.join(self.project_dir(), "dodo_commands", "res")

    def package_dir(self):
        import dodo_commands
        return os.path.dirname(dodo_commands.__file__)

    def extra_dir(self):
        return os.path.join(self.package_dir(), "extra")



from argparse import ArgumentParser
from dodo_commands.framework import Dodo
from dodo_commands.framework.config import expand_keys


def _args():
    parser = ArgumentParser()
    parser.add_argument('text')
    args = Dodo.parse_args(parser)
    return args


if Dodo.is_main(__name__):
    args = _args()
    print(expand_keys(Dodo.config, args.text))



import argcomplete
from argparse import ArgumentParser
from collections import defaultdict
import ruamel.yaml
import os
import pkgutil
import subprocess
import sys
import traceback
from importlib import import_module
from dodo_commands.framework.util import query_yes_no
from dodo_commands.framework.config import (CommandPath, ConfigLoader,
                                            load_global_config_parser)
from dodo_commands.framework.paths import Paths
from dodo_commands.framework.singleton import Dodo
from dodo_commands.framework.command_error import CommandError  # noqa


def get_version():  # noqa
    return "0.16.6"


def execute_script(package_path, command_name):
    """
    Executes the script associated with command_name by importing its package.
    The script is assumed to have an entry point that is executed if
    Dodo.is_main(__name__) is True.
    """

    def install_packages(meta_data_filename):
        """Pip install packages found in meta_data_filename."""
        with open(meta_data_filename) as f:
            meta_data = ruamel.yaml.round_trip_load(f.read())
            print("This command wants to install additional packages:\n")
            print(meta_data['requirements'])
            if query_yes_no("Install (yes), or abort (no)?"):
                print("\n--- Installing from %s ---" % meta_data_filename)
                pip = Paths().pip()
                subprocess.check_call([pip, "install"] +
                                      meta_data['requirements'])
                print("--- Done ---\n\n")
            else:
                sys.exit(1)

    import_path = '%s.%s' % (package_path, command_name)
    if package_path in ("", None, "."):
        import_path = command_name

    try:
        import_module(import_path)
    except ImportError as e:
        try:
            base_path = import_module(package_path).__path__[0]
            meta_data_filename = os.path.join(base_path,
                                              command_name + ".meta")
            if os.path.exists(meta_data_filename):
                install_packages(meta_data_filename)
                import_module(import_path)
            else:
                print(traceback.print_exc(e))
                sys.exit(1)
        except ImportError as e:
            print(traceback.print_exc(e))
            sys.exit(1)


def get_command_map():
    """
    Return a dictionary mapping command names to their Python module directory.
    The dictionary is in the format {command_name: module_name}.
    """
    command_map = {}
    command_path = CommandPath(ConfigLoader().load())
    command_path.extend_sys_path()
    for item in command_path.items:
        commands = [
            name for _, name, is_pkg in pkgutil.iter_modules([item])
            if not is_pkg and not name.startswith('_')
        ]
        for command_name in commands:
            command_map[command_name] = os.path.basename(item)
    return command_map


class ManagementUtility(object):
    """Internal helper class for executing commands."""

    def __init__(self, argv):  # noqa
        self.argv = argv
        self.prog_name = os.path.basename(self.argv[0])

    def main_help_text(self, commands_only=False, command_map=None):
        """Return the script's main help text, as a string."""
        if command_map is None:
            command_map = get_command_map()

        if commands_only:
            usage = sorted(command_map.keys())
        else:
            usage = [
                "",
                "Version %s (%s --version)." % (get_version(), self.prog_name),
                "Type '%s help <command>' for help on "
                "a specific command." % self.prog_name,
                "Available commands (dodo help --commands):",
            ]
            command_groups = defaultdict(lambda: [])
            for command_name, package_path in command_map.items():
                command_groups[package_path].append(command_name)
            for package_path in sorted(command_groups.keys()):
                usage.append("")
                for command_name in sorted(command_groups[package_path]):
                    usage.append("    %s" % command_name)

        return '\n'.join(usage)

    def _handle_exception(self, e):
        if (getattr(Dodo.args, 'traceback', False)
                or not isinstance(e, CommandError)):
            raise
        sys.stderr.write('%s: %s\n' % (e.__class__.__name__, e))
        sys.exit(1)

    def execute(self):
        """
        Execute command.

        Given the command-line arguments, this figures out which command is
        being run, creates a parser appropriate to that command, and runs it.
        """
        try:
            command_map = get_command_map()
        except Exception as e:
            self._handle_exception(e)

        if "_ARGCOMPLETE" in os.environ:
            words = os.environ['COMP_LINE'].split()
            command_name = words[1]

            if command_name not in command_map:
                parser = ArgumentParser()
                parser.add_argument(
                    'command', choices=[x for x in command_map.keys()])
                argcomplete.autocomplete(parser)

            os.environ['COMP_LINE'] = ' '.join(words[:1] + words[2:])
            os.environ['COMP_POINT'] = str(
                int(os.environ['COMP_POINT']) - (len(command_name) + 1))
        else:
            try:
                command_name = self.argv[1]
            except IndexError:
                command_name = 'help'  # Display help if no arguments were given.

        global_config = load_global_config_parser()
        if global_config.has_section('alias'):
            for key, val in global_config.items('alias'):
                if key == command_name:
                    command_name = val

        if command_name == '--version':
            sys.stdout.write(get_version() + '\n')
        elif command_name == 'help':
            if '--commands' in sys.argv:
                sys.stdout.write(
                    self.main_help_text(
                        commands_only=True, command_map=command_map) + '\n')
            else:
                sys.stdout.write(self.main_help_text() + '\n')
        else:
            if command_name not in command_map:
                print("Unknown dodo command: %s" % command_name)
                sys.exit(1)

            package_path = command_map[command_name]
            Dodo.command_name = command_name
            Dodo.package_path = package_path

            try:
                execute_script(package_path, command_name)
            except KeyboardInterrupt:
                print('\n')
                sys.exit(1)
            except Exception as e:
                self._handle_exception(e)


def execute_from_command_line(argv):
    """A simple method that runs a ManagementUtility."""
    utility = ManagementUtility(argv)
    utility.execute()



from argparse import ArgumentParser
from dodo_commands.framework import Dodo
from dodo_commands.framework.paths import Paths
from dodo_commands.framework.config import (CommandPath,
                                            load_global_config_parser)
import os
import sys


def _args():  # noqa
    parser = ArgumentParser(
        description=('Print the value of /ROOT/<what>_dir. For example: ' +
                     '"dodo which src" prints the value of /ROOT/src_dir.'))
    parser.add_argument(
        'what',
        nargs='?',
    )

    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        '--config', action="store_true", help='Print where the config file is')
    group.add_argument(
        '--global-config',
        action="store_true",
        help='Print where the global config file is')
    group.add_argument(
        '--script',
        help='Print where the dodo command script with given name is')
    group.add_argument(
        '--dir',
        dest='directory',
        help=
        'Finds the X_dir value in the configuration, where X is the given option value'
    )
    group.add_argument(
        '--decorators',
        action="store_true",
        help='Prints which command decorators are available')
    group.add_argument(
        '--projects',
        action="store_true",
        help='Prints which projects are available')
    group.add_argument(
        '--default-commands',
        action="store_true",
        help='Prints the directory where default commands are stored')

    return Dodo.parse_args(parser)


def _which_script(script):
    command_path = CommandPath(Dodo.config)
    for item in command_path.items:
        script_path = os.path.join(item, script + ".py")
        if os.path.exists(script_path):
            return script_path
    return None


def _which_dir(directory):
    return Dodo.get_config("/ROOT/%s_dir" % directory, None)


def _projects_dir():
    return os.path.expanduser(load_global_config_parser().get(
        "settings", "projects_dir"))


if Dodo.is_main(__name__):
    args = _args()

    if args.config:
        print(os.path.join(Paths().res_dir(), "config.yaml"))
    elif args.global_config:
        sys.stdout.write(Paths().global_config_filename() + '\n')
    elif args.default_commands:
        sys.stdout.write(Paths().default_commands_dir() + '\n')
    elif args.script:
        sys.stdout.write(_which_script(args.script) + '\n')
    elif args.directory:
        sys.stdout.write(_which_dir(args.directory) + '\n')
    elif args.decorators:
        sys.stdout.write(", ".join(sorted(Dodo.all_decorators().keys())) +
                         '\n')
    elif args.projects:
        projects = [
            x for x in os.listdir(_projects_dir()) if os.path.isdir(
                os.path.join(_projects_dir(), x, "dodo_commands", "res"))
        ]
        sys.stdout.write('\n'.join(sorted(projects)) + '\n')
    elif args.what:
        x = _which_script(args.what) or _which_dir(args.what)
        if x:
            sys.stdout.write(x + '\n')
    else:
        sys.stdout.write(Dodo.get_config("/ROOT/project_name") + '\n')



from argparse import ArgumentParser
from dodo_commands.framework import Dodo


def _args():  # noqa
    parser = ArgumentParser(description=('Print hello world.'))
    args = Dodo.parse_args(parser)
    return args


if Dodo.is_main(__name__):
    args = _args()
    Dodo.runcmd(['echo', 'Hello world'])



from dodo_commands.framework.args_tree import ArgsTreeNode
from dodo_commands.framework import Dodo


class Decorator:  # noqa
    def add_arguments(self, parser):  # noqa
        parser.add_argument(
            '--use-debugger',
            action='store_true',
            default=False,
            help="Run the command through the debugger")

    def modify_args(self, root_node, cwd):  # noqa
        if not Dodo.args.use_debugger:
            return root_node, cwd

        debugger_node = ArgsTreeNode(
            "debugger", args=[Dodo.get_config('/BUILD/debugger')])
        debugger_node.add_child(root_node)
        return debugger_node, cwd



from argparse import ArgumentParser
from dodo_commands.framework import Dodo, CommandError
from dodo_commands.framework.paths import Paths
from dodo_commands.framework.util import is_using_system_dodo, symlink
import sys
import os


def _args():
    parser = ArgumentParser(
        description=("Install default command directories supplied by the " +
                     "paths argument: dodo install-default-commands " +
                     "/path/to/my/commands. " + _packages_in_extra_dir()))
    parser.add_argument(
        "paths",
        nargs='*',
        help='Create symlinks to these command directories')
    parser.add_argument(
        "--pip", nargs='*', help='Pip install the commands in these packages')
    parser.add_argument(
        "--remove",
        action='store_true',
        help='Remove commands from the default commands directory')
    args = Dodo.parse_args(parser)
    return args


def _packages_in_extra_dir():
    extra_dir = Paths().extra_dir()
    packages = [
        x for x in os.listdir(extra_dir)
        if os.path.isdir(os.path.join(extra_dir, x)) and not x.startswith('__')
    ]

    if len(packages) == 0:
        return ""

    if len(packages) == 1:
        msg = " The %s package is found automagically " % packages[0]
    else:
        packages[-1] = "and " + packages[-1]
        msg = " The %s packages are found automagically " % ", ".join(packages)

    return (msg + " in the dodo_commands.extra package" +
            ", e.g. the following works: dodo install-default-commands %s." %
            packages[0])


def _report_error(msg):
    sys.stderr.write(msg + os.linesep)


def _remove_commands(path):
    """Install the dir with the default commands."""
    dest_dir = os.path.join(Paths().default_commands_dir(),
                            os.path.basename(path))

    if not os.path.exists(dest_dir):
        raise CommandError("Directory does not exist: %s" % dest_dir)
    Dodo.runcmd(['rm', '-rf', dest_dir])


def _install_commands(path):
    """Install the dir with the default commands."""
    dest_dir = os.path.join(Paths().default_commands_dir(),
                            os.path.basename(path))

    if not os.path.exists(path):
        alt_path = os.path.join(Paths().extra_dir(), path)
        if os.path.exists(alt_path):
            path = alt_path
        else:
            _report_error("Error: path not found: %s" % path)
            return False

    if os.path.exists(dest_dir):
        return False

    try:
        if os.name == 'nt' and not args.confirm:
            symlink(os.path.abspath(path), dest_dir, target_is_directory=True)
        else:
            Dodo.runcmd(['ln', '-s', os.path.abspath(path), dest_dir])
    except:
        _report_error("Error: could not create a symlink in %s." % dest_dir)
        return False

    return True


def _remove_package(package):
    """Install the dir with the default commands."""
    dest_dir = os.path.join(Paths().default_commands_dir(), package)

    if not os.path.exists(dest_dir):
        raise CommandError("Directory does not exist: %s" % dest_dir)
    Dodo.runcmd(['rm', '-rf', dest_dir])


def _install_package(package):
    pip = Paths().pip()
    if pip.startswith('/usr/') and not os.path.exists(pip):
        alt_pip = pip.replace("/usr/", "/usr/local/")
        if os.path.exists(alt_pip):
            pip = alt_pip
        else:
            raise CommandError(
                "Expected to find a pip executable at location %s or %s." %
                (pip, alt_pip))

    Dodo.runcmd([
        pip, 'install', '--upgrade', '--target',
        Paths().default_commands_dir(), package
    ])


if Dodo.is_main(__name__):
    args = _args()
    if args.pip and not is_using_system_dodo():
        raise CommandError(
            "Please deactivate your dodo project first by running 'deactivate'."
        )

    if args.paths:
        for path in args.paths:
            if args.remove:
                _remove_commands(path)
            else:
                _install_commands(path)

    if args.pip:
        for package in args.pip:
            if args.remove:
                _remove_package(package)
            else:
                _install_package(package)



from argparse import ArgumentParser
from dodo_commands.framework import Dodo
from dodo_commands.framework.util import bordered
from jinja2 import Environment, FileSystemLoader
import glob
import os
import sphinx  # noqa
import sys
from importlib import import_module


def _args():
    parser = ArgumentParser(
        description='Create documentation based on the dodo config')
    args = Dodo.parse_args(parser)
    args.src_dir = Dodo.get_config('/DIAGNOSE/src_dir')
    args.output_dir = Dodo.get_config('/DIAGNOSE/output_dir')
    args.project_name = Dodo.get_config('/ROOT/project_name')
    args.filters = Dodo.get_config('/DIAGNOSE/filters', [])
    return args


def _template_files(args):
    result = []
    for step_file in glob.glob(os.path.join(args.src_dir, "*.rst")):
        template_file = os.path.basename(step_file)
        result.append(template_file)
    return result


def _create_sphinx_conf(args):
    target_conf_file = os.path.join(args.output_dir, 'conf.py')
    src_conf_file = os.path.join(args.src_dir, 'conf.py')
    if os.path.exists(src_conf_file):
        Dodo.runcmd(['cp', src_conf_file, target_conf_file])
    elif not os.path.exists(target_conf_file):
        Dodo.runcmd(
            ['sphinx-quickstart',
             '--project=%s' % args.project_name],
            cwd=args.output_dir)


def _sphinx_build(args):
    Dodo.runcmd([
        'sphinx-build',
        '-q',
        '-b',
        'html',
        args.output_dir,
        os.path.join(args.output_dir, 'html'),
    ], )


def _open_browser(args):
    Dodo.runcmd([
        'xdg-open',
        os.path.join(args.output_dir, 'html', 'index.html'),
    ], )


def _create_jinja_environment(args):
    env = Environment(
        loader=FileSystemLoader(args.src_dir),
        trim_blocks=True,
        lstrip_blocks=True,
    )

    # import more filters and tests
    for syspath, basename in args.filters:
        sys.path.append(syspath)
        import_module(basename)
        sys.path.pop()

    env.tests.update(_tests())
    env.filters.update(_filters())
    return env


def _report_error(errors, error):
    errors.append(error)


def _tests():
    result = {}
    from dodo_standard_commands._diagnose_filters_and_tests import tests
    for name, func in tests.items():
        result[name] = func
    return result


def _filters():
    result = {}
    from dodo_standard_commands._diagnose_filters_and_tests import filters
    for name, func in filters.items():
        result[name] = func
    return result


def _render_template(args, env, template_file):
    template = env.get_template(template_file)
    output = template.render()
    rst_file = os.path.join(args.output_dir, template_file)
    with open(rst_file, 'w') as ofs:
        ofs.write(output)


if Dodo.is_main(__name__, safe=False):
    args = _args()

    errors = []
    if not os.path.exists(args.output_dir):
        Dodo.runcmd(['mkdir', '-p', args.output_dir])

    env = _create_jinja_environment(args)

    template_files = _template_files(args)
    for template_file in template_files:
        _render_template(args, env, template_file)

    if errors:
        print("")
        print(bordered("Warning, there were errors:") + "\n" +
              ("\n".join(errors)))
        print("")
    _create_sphinx_conf(args)
    _sphinx_build(args)
    _open_browser(args)



"""Module for working with the dodo configurations."""

from dodo_commands.framework.command_error import CommandError
from dodo_commands.framework.config_expander import ConfigExpander
from dodo_commands.framework.config_expander import Key, KeyNotFound  # noqa
from dodo_commands.framework.paths import Paths
from dodo_commands.framework.util import symlink
from six.moves import configparser
import glob
import json
import os
import re
import ruamel.yaml
import sys
import hashlib


def load_global_config_parser():
    config_parser = configparser.ConfigParser()
    config_parser.read(Paths().global_config_filename())
    return config_parser


def write_global_config_parser(config_parser):
    """Save configuration."""
    with open(Paths().global_config_filename(), "w") as f:
        config_parser.write(f)


_global_config = """[settings]
projects_dir=~/projects
python_interpreter=python
diff_tool=diff

[recent]
"""


def create_global_config():
    """Create config file and default_commands dir."""
    base_dir = Paths().global_config_dir()
    if not os.path.exists(base_dir):
        os.mkdir(base_dir)

    config_filename = Paths().global_config_filename()
    if not os.path.exists(config_filename):
        with open(config_filename, 'w') as f:
            f.write(_global_config)

    default_commands_dir = Paths().default_commands_dir()
    if not os.path.exists(default_commands_dir):
        os.mkdir(default_commands_dir)
        symlink(
            os.path.join(Paths().extra_dir(), "dodo_standard_commands"),
            os.path.join(Paths().default_commands_dir(),
                         "dodo_standard_commands"),
            target_is_directory=True)

    init_py = os.path.join(default_commands_dir, "__init__.py")
    if not os.path.exists(init_py):
        with open(init_py, 'w') as f:
            pass


def merge_into_config(config, layer, xpath=None):
    def _is_list(x):
        return isinstance(x, type(list()))

    def _is_dict(x):
        return isinstance(x, type(dict()))

    def _raise(xpath):
        raise CommandError(
            "Cannot merge configurations. Check key /%s" % '/'.join(new_xpath))

    xpath = xpath or []
    for key, val in (layer or {}).items():
        new_xpath = xpath + [key]

        if key not in config:
            config[key] = val
        elif _is_dict(val):
            if not _is_dict(config[key]):
                _raise(new_xpath)
            merge_into_config(config[key], val, new_xpath)
        elif _is_list(val):
            if not _is_list(config[key]):
                _raise(new_xpath)
            config[key].extend(val)
        else:
            config[key] = val


class ConfigIO:
    """Read and write the dodo configuration."""

    def __init__(self, config_base_dir=None):
        """Arg config_base_dir is where config files are searched.

        Arg config_base_dir defaults to Paths().res_dir().
        """
        self.config_base_dir = (Paths().res_dir() if config_base_dir is None
                                else config_base_dir)  # noqa

    def _path_to(self, post_fix_paths):
        """Return path composed of config_base_dir and post_fix_paths list."""
        return os.path.join(self.config_base_dir, *post_fix_paths)

    def get_layers(self, config):
        """Returns list of layer filenames"""

        def add_prefix(filename):
            return (filename
                    if filename.startswith('/') else self._path_to([filename]))

        patterns = [
            add_prefix(os.path.expanduser(pattern))
            for pattern in config.get('ROOT', {}).get('layers', [])
        ]

        result = []
        for pattern in patterns:
            result.extend([x for x in glob.glob(pattern)])
        return result

    def _load_layers(self, config):
        def layer_exists(layer_filename):
            if not os.path.exists(layer_filename):
                print("Warning: layer not found: %s" % layer_filename)
                return False
            return True

        layer_filenames = [
            x for x in self.get_layers(config) if layer_exists(x)
        ]
        layers = [self.load(x, load_layers=False) for x in layer_filenames]
        for layer in layers:
            merge_into_config(config, layer)

    def load(self, config_filename='config.yaml', load_layers=True):
        """Get configuration."""
        full_config_filename = self._path_to([config_filename])
        if not os.path.exists(full_config_filename):
            return None

        with open(full_config_filename) as f:
            config = ruamel.yaml.round_trip_load(f.read())
        if load_layers:
            self._load_layers(config)
        return config

    def save(self, config, config_filename='config.yaml'):
        """Write config to config_filename."""
        with open(self._path_to([config_filename]), 'w') as f:
            return f.write(ruamel.yaml.round_trip_dump(config))


class ConfigLoader:
    """Load the project's dodo config and expand it."""

    def _add_to_config(self, config, section, key, value):
        if section in config:
            if key not in config[section]:
                config[section][key] = value

    def _system_commands_dir(self):
        """Return directory where system command scripts are stored"""
        import dodo_commands.dodo_system_commands
        return os.path.dirname(dodo_commands.dodo_system_commands.__file__)

    def _extend_config(self, config):
        """Add special values to the project's config"""
        project_dir = Paths().project_dir()
        if project_dir:
            self._add_to_config(config, "ROOT", "project_name",
                                os.path.basename(project_dir))
            self._add_to_config(config, "ROOT", "project_dir", project_dir)
            self._add_to_config(config, "ROOT", "res_dir", Paths().res_dir())

    def _extend_command_path(self, config):
        """Add the system commands to the command path"""
        self._add_to_config(config, "ROOT", "command_path", [])
        config['ROOT']['command_path'].append(self._system_commands_dir())

    def _report(self, x):
        sys.stderr.write(x)
        sys.stderr.flush()

    def load(self, config_base_dir=None):
        fallback_config = dict(ROOT={})
        try:
            config = ConfigIO(config_base_dir).load() or fallback_config
        except ruamel.yaml.scanner.ScannerError:
            config = fallback_config
            self._report(
                "There was an error while loading the configuration. "
                "Run 'dodo diff' to compare your configuration to the "
                "default one.\n")

        self._extend_command_path(config)
        self._extend_config(config)
        ConfigExpander().run(config)
        return config


class CommandPath:
    """Read search paths for command scripts from the configuration."""

    def __init__(self, config):
        exclude_patterns = self._exclude_patterns(config)
        excluded_command_dirs = self._create_items(exclude_patterns)

        include_patterns = self._include_patterns(config)
        self.items = sorted([
            x for x in self._create_items(include_patterns)
            if x not in excluded_command_dirs
        ])

        basenames = [os.path.basename(x) for x in self.items]
        for basename in basenames:
            if basenames.count(basename) > 1:
                raise CommandError(
                    "More than 1 command path with name %s" % basename)

    def _include_patterns(self, config):
        return config.get('ROOT', {}).get('command_path', [])

    def _exclude_patterns(self, config):
        return config.get('ROOT', {}).get('command_path_exclude', [])

    def _create_items(self, patterns):
        result = []
        for pattern in patterns:
            for x in glob.glob(os.path.expanduser(pattern)):
                if os.path.isdir(x):
                    result.append(x)
        return result

    def _create_search_path_dir(self):
        hash_code = hashlib.md5(json.dumps(
            self.items).encode('utf-8')).hexdigest()
        search_path_dir = os.path.join(Paths().global_config_dir(),
                                       "search_path", hash_code)

        if not os.path.exists(search_path_dir):
            os.makedirs(search_path_dir)
            open(os.path.join(search_path_dir, "__init__.py"), 'a').close()
            for item in self.items:
                basename = os.path.basename(item)
                symlink(
                    item,
                    os.path.join(search_path_dir, basename),
                    target_is_directory=True)

        return search_path_dir

    def extend_sys_path(self):  # noqa
        search_path_dir = self._create_search_path_dir()
        if search_path_dir not in sys.path:
            sys.path.append(search_path_dir)


def look_up_key(config, key, default_value="__not_set_234234__"):
    xpath = [k for k in key.split("/") if k]
    try:
        return Key(config, xpath).get()
    except KeyNotFound:
        if default_value == "__not_set_234234__":
            raise
    return default_value


def expand_keys(config, text):
    result = ""
    val_terms = re.split('\$\{([^\}]+)\}', text)
    for idx, term in enumerate(val_terms):
        if idx % 2:
            str_rep = json.dumps(look_up_key(config, term))
            if str_rep.startswith('"') and str_rep.endswith('"'):
                str_rep = str_rep[1:-1]
            result += str_rep
        else:
            result += term
    return result



from argparse import ArgumentParser
from dodo_commands.framework import Dodo, CommandError
from dodo_commands.framework.paths import Paths
from dodo_commands.framework.config import (load_global_config_parser,
                                            ConfigIO)
from six.moves import configparser
import os


def _args():
    parser = ArgumentParser()
    args = Dodo.parse_args(parser)
    args.editor = load_global_config_parser().get("settings", "config_editor")
    args.res_dir = Dodo.get_config('/ROOT/res_dir')
    return args


if Dodo.is_main(__name__, safe=True):
    try:
        args = _args()
    except configparser.NoOptionError as e:
        raise CommandError("{error}. Please check {filename}".format(
            error=str(e), filename=Paths().global_config_filename()))

    yaml_filenames = ([os.path.join(args.res_dir, 'config.yaml')] +
                      ConfigIO().get_layers(Dodo.config))

    yaml_filenames.append(Paths().global_config_filename())
    Dodo.runcmd([args.editor] + yaml_filenames, cwd='.')



from argparse import ArgumentParser
from dodo_commands.framework import Dodo
from plumbum.cmd import docker
from six.moves import input as raw_input
import sys


def _args():
    parser = ArgumentParser()
    parser.add_argument('name', nargs='?')
    args = Dodo.parse_args(parser)
    return args


if Dodo.is_main(__name__):
    args = _args()
    if not args.name:
        containers = [
            x for x in docker("ps", "--format", "{{.Names}}").split('\n') if x
        ]
        print("0 - exit")
        for idx, container in enumerate(containers):
            print("%d - %s" % (idx + 1, container))

        print("\nSelect a container: ")
        choice = int(raw_input()) - 1

        if choice == -1:
            sys.exit(0)

        args.name = containers[choice]

    Dodo.runcmd([
        'docker',
        'exec',
        '-i',
        '-t',
        args.name,
        '/bin/bash',
    ], )



"""Pull the latest version of the Dodo Commands system."""
from argparse import ArgumentParser
from dodo_commands.framework import Dodo, CommandError
from dodo_commands.framework.config import ConfigIO
from dodo_commands.framework.util import symlink
import glob
import os
import sys


def _args():
    parser = ArgumentParser()
    parser.add_argument(
        'src_dir', help="The src directory for the bootstrapped project")
    parser.add_argument(
        'shared_config_dir',
        help=
        "Location relative to src_dir where the shared project config is stored",
    )
    parser.add_argument('--force', dest="use_force", action="store_true")

    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        '--git-url',
        dest='git_url',
        help="Clone this repository to the src_dir location",
    )
    group.add_argument(
        '--link-dir',
        help="Make the src directory a symlink to this directory",
    )
    group.add_argument(
        '--cookiecutter-url',
        help="Use cookiecutter to create the src_dir location",
    )

    parser.add_argument(
        '--depth', type=int, default=0, help="Depth for cloning repositories")
    parser.add_argument('--branch', help="Branch to checkout after cloning")

    parser.add_argument(
        '--src-subdir',
        help="Specify a subdirectory of src_dir to clone into",
    )

    args = Dodo.parse_args(parser)
    args.project_dir = Dodo.get_config('/ROOT/project_dir')
    return args


def _report(msg):
    sys.stderr.write(msg + "\n")


def _copy_defaults(args, shared_config_dir):
    res_dir = os.path.join(args.project_dir, "dodo_commands", "res")
    for filename in glob.glob(os.path.join(shared_config_dir, "*")):
        dest_path = os.path.join(res_dir, os.path.basename(filename))
        if os.path.exists(dest_path):
            if args.confirm:
                print(
                    "Warning, destination path already exists: %s" % dest_path)
            elif args.use_force:
                print("Overwriting existing path: %s" % dest_path)
            else:
                raise CommandError(
                    "Destination path %s already exists. " % dest_path +
                    "Use the --confirm or --force flag to overwrite it.")
        Dodo.runcmd(["cp", "-rf", filename, dest_path])


def _clone(args, src_dir):
    if os.path.exists(src_dir):
        raise CommandError(
            "Cannot clone into %s, path already exists" % src_dir)

    Dodo.runcmd(
        ["git", "clone", args.git_url,
         os.path.basename(src_dir)] + (["--depth", args.depth]
                                       if args.depth else []),
        cwd=os.path.dirname(src_dir))
    if args.branch:
        Dodo.runcmd(["git", "checkout", args.branch], cwd=src_dir)


def _cookiecutter(src_dir, cookiecutter_url):
    if os.path.exists(src_dir):
        raise CommandError(
            "Cannot clone into %s, path already exists" % src_dir)

    Dodo.runcmd(["cookiecutter", cookiecutter_url, "-o", src_dir])


def _link_dir(link_target, link_name):
    if os.path.exists(link_name):
        raise CommandError(
            "Cannot create a link because %s already exists" % link_name)
    if os.name == 'nt' and not args.confirm:
        symlink(link_target, link_name, target_is_directory=True)
    else:
        Dodo.runcmd(["ln", "-s", link_target, link_name])


def _update_config(src_dir, relative_shared_config_dir):
    config = ConfigIO().load(load_layers=False)
    config['ROOT']['src_dir'] = src_dir
    config['ROOT'][
        'shared_config_dir'] = '${/ROOT/src_dir}/%s' % relative_shared_config_dir
    ConfigIO().save(config)


def _get_full_src_dir(src_dir, src_subdir):
    postfix = (os.path.join(src_dir, src_subdir) if src_subdir else src_dir)
    return (postfix if os.path.isabs(src_dir) else os.path.join(
        args.project_dir, postfix))


if Dodo.is_main(__name__, safe=False):
    args = _args()

    full_src_dir = _get_full_src_dir(args.src_dir, args.src_subdir)
    Dodo.runcmd(['mkdir', '-p', os.path.dirname(full_src_dir)])

    if args.git_url:
        _clone(args, full_src_dir)
    elif args.link_dir:
        _link_dir(os.path.expanduser(args.link_dir), full_src_dir)
    elif args.cookiecutter_url:
        _cookiecutter(full_src_dir, os.path.expanduser(args.cookiecutter_url))

    shared_config_dir = os.path.join(full_src_dir, args.shared_config_dir)
    if not os.path.exists(shared_config_dir):
        raise CommandError(
            "Shared config directory %s not found." % shared_config_dir)

    _copy_defaults(args, shared_config_dir)
    _update_config(
        args.src_dir if os.path.isabs(args.src_dir) else os.path.join(
            "${/ROOT/project_dir}", args.src_dir),
        os.path.join(args.src_subdir or '', args.shared_config_dir))



from dodo_commands.framework.util import classproperty, query_yes_no
from importlib import import_module
import argcomplete
import argparse
import fnmatch
import os
import sys
from dodo_commands.framework.config import (ConfigLoader, look_up_key,
                                            CommandPath)
from dodo_commands.framework.args_tree import ArgsTreeNode
from dodo_commands.framework.command_error import CommandError
from plumbum import FG, ProcessExecutionError, local


def _ask_to_continue(args, cwd, is_echo, is_confirm):
    """Ask the user whether to continue with executing func."""

    def to_str():
        return args.to_str(slash=args.children)

    if is_echo:
        print(to_str())
        return False

    if is_confirm:
        print("(%s) %s" % (cwd, to_str()))
        if not query_yes_no("continue?"):
            return False
        else:
            print("")

    return True


class Dodo:
    command_name = None
    safe = True
    args = argparse.Namespace()

    _config = None

    @classproperty
    def config(cls):  # noqa
        if cls._config is None:
            cls._config = ConfigLoader().load()
        return cls._config

    @classmethod
    def create_get_config(cls, local_vars):
        """
        Returns a get_config function that interpolates local_vars.
        For example:

        gc = Dodo.create_get_config(dict(user='sam'))
        tmp_dir = gc('/USERS/{user}/tmp_dir', default='/tmp')
        """

        def get_config(key, default=None):
            return cls.get_config(key.format(**local_vars), default)

        return get_config

    @classmethod
    def get_config(cls, key, default_value="__not_set_234234__"):  # noqa
        return look_up_key(cls.config, key, default_value)

    @classmethod
    def is_main(cls, name, safe=None):
        if safe is not None:
            cls.safe = safe
        return name in ('__main__', cls.package_path + '.' + cls.command_name)

    @classmethod
    def _get_decorators(cls):
        return [
            cls._load_decorator(name, directory)
            for name, directory in cls.all_decorators().items()
            if cls._uses_decorator(name)
        ]

    @classmethod
    def _uses_decorator(cls, decorator_name):
        patterns = (cls.config.get('ROOT', {}).get('decorators', {}).get(
            decorator_name, []))
        command_name = cls.command_name
        approved = [
            pattern for pattern in patterns if not pattern.startswith("!")
            and fnmatch.filter([command_name], pattern)
        ]
        rejected = [
            pattern for pattern in patterns if pattern.startswith("!")
            and fnmatch.filter([command_name], pattern[1:])
        ]
        return len(approved) and not len(rejected)

    @classmethod
    def _load_decorator(cls, name, directory):
        """Load and return decorator class in module with given name."""
        return import_module(directory + "." + name).Decorator()

    @classmethod
    def all_decorators(cls):
        """Returns a mapping from decorator name to its directory."""
        command_path = CommandPath(cls.config)
        command_path.extend_sys_path()
        result = {}
        for item in command_path.items:
            try:
                module_path = os.path.basename(item) + ".decorators"
                module = import_module(module_path)
                for decorator in os.listdir(module.__path__[0]):
                    name, ext = os.path.splitext(decorator)
                    if ext == '.py' and name != '__init__':
                        result[name] = module_path
            except ImportError:
                continue
        return result

    @classmethod
    def parse_args(cls, parser):
        if not cls.command_name:
            argcomplete.autocomplete(parser)
            return parser.parse_args(sys.argv)

        parser.add_argument(
            '--traceback', action='store_true', help=argparse.SUPPRESS)

        parser.add_argument(
            '--confirm',
            action='store_true',
            help="Confirm each performed action before its execution")

        parser.add_argument(
            '--echo',
            action='store_true',
            help="Print all commands instead of running them")

        for decorator in cls._get_decorators():
            decorator.add_arguments(parser)

        argcomplete.autocomplete(parser)
        parser.prog = "%s %s" % (os.path.basename(sys.argv[0]), sys.argv[1])
        cls.args = parser.parse_args(sys.argv[2:])

        if cls.args.echo and not cls.safe:
            raise CommandError(
                "The --echo option is not supported for unsafe commands.\n" +
                "Since this command is marked as unsafe, some operations will "
                + "be performed directly, instead of echoed to the console. " +
                "Use --confirm if you wish to execute with visual feedback. ")

        if cls.args.confirm and not cls.safe:
            if not query_yes_no(
                    "Since this command is marked as unsafe, some operations will "
                    +
                    "be performed without asking for confirmation. Continue?",
                    default="no"):
                return

        return cls.args

    @classmethod
    def runcmd(cls, args, cwd=None, quiet=False, capture=False):
        if not hasattr(cls.args, 'echo'):
            raise CommandError('Dodo.runcmd was called without first calling '
                               'Dodo.parse_args.')
        root_node = ArgsTreeNode('original_args', args=args)
        for decorator in cls._get_decorators():
            root_node, cwd = decorator.modify_args(root_node, cwd)

        if not _ask_to_continue(root_node, cwd or local.cwd, cls.args.echo,
                                cls.args.confirm):
            return False

        if cwd and not os.path.exists(cwd):
            raise CommandError("Directory not found: %s" % cwd)

        with local.cwd(cwd or local.cwd):
            flat_args = root_node.flatten()
            func = local[flat_args[0]][flat_args[1:]]
            variable_map = cls.get_config('/ENVIRONMENT/variable_map', {})
            with local.env(**variable_map):
                if capture:
                    return func()
                try:
                    func & FG
                    return True
                except ProcessExecutionError:
                    if not quiet:
                        print(
                            "\nDodo Commands error while running this command:"
                        )
                        print("\n\n%s" % func)
                    return False
